"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.createApp = void 0;
const picocolors_1 = __importDefault(require("picocolors"));
const fs_1 = __importDefault(require("fs"));
const metalsmith_1 = __importDefault(require("metalsmith"));
const inquirer_1 = __importDefault(require("inquirer"));
const nanospinner_1 = require("nanospinner");
const consolidate_1 = require("consolidate");
const path_1 = __importDefault(require("path"));
const logger_1 = __importDefault(require("../logger"));
const joycon_1 = __importDefault(require("joycon"));
const utils_1 = require("../utils");
const tmp_1 = __importDefault(require("tmp"));
const utils_2 = require("../utils");
const render = consolidate_1.ejs.render;
async function getOptions(projectTemplatePath) {
    var _a;
    const joycon = new joycon_1.default({ cwd: projectTemplatePath });
    const moduleName = "meta";
    const config = await joycon.load([
        `.${moduleName}rc`,
        `.${moduleName}.json`,
        `.${moduleName}.yaml`,
        `.${moduleName}.yml`,
        `.${moduleName}.js`,
        `${moduleName}.js`,
    ]);
    return (_a = config.data) !== null && _a !== void 0 ? _a : {};
}
function copyTo(src, dest, done) {
    var metalsmith = (0, metalsmith_1.default)(src);
    metalsmith
        .clean(false)
        .source(".")
        .destination(dest)
        .build((err) => {
        if (err)
            throw err;
        done();
    });
}
async function initProject(config) {
    const metaOpts = await getOptions(config.projectTemplatePath);
    runMetalsmith(config, metaOpts);
}
function runMetalsmith(config, metaOpts) {
    var _a;
    const metalsmith = (0, metalsmith_1.default)(path_1.default.join(config.projectTemplatePath, "template"));
    const metaData = metalsmith.metadata();
    const questions = metaOpts && metaOpts.questions;
    config.data = (_a = config.data) !== null && _a !== void 0 ? _a : {};
    metaData.destPath = config.destPath
        ? path_1.default.resolve(config.destPath)
        : path_1.default.join(process.cwd(), config.data.projectName || "");
    metalsmith.use(askQuestions(questions, config)).use(renderTemplateFiles());
    metalsmith
        .clean(false)
        .source(".")
        .destination(metaData.destPath)
        .build((err, files) => {
        if (err)
            throw err;
        if (typeof metaOpts.endCallback === "function") {
            const helpers = { chalk: picocolors_1.default, logger: logger_1.default, files };
            metaData.answers = metaData.data;
            metaOpts.endCallback(metaData, helpers);
        }
        else {
            logger_1.default.success("init success");
        }
    });
}
function askQuestions(questions, config) {
    return (files, metalsmith, done) => {
        var metadata = metalsmith.metadata();
        new Promise((resolve) => {
            if (config.shouldAskQuestions) {
                inquirer_1.default.prompt(questions).then(resolve);
            }
            else if (!questions || !questions.length) {
                resolve({});
            }
            else {
                resolve(config.data);
            }
        }).then((data) => {
            metadata.data = { ...config.data, ...data };
            done(null, files, metalsmith);
        });
    };
}
function renderTemplateFiles() {
    return (files, metalsmith, done) => {
        const keys = Object.keys(files);
        const metaData = metalsmith.metadata();
        const promises = keys.map((key) => {
            return new Promise((resolve, reject) => {
                const str = files[key].contents.toString();
                render(str, metaData.data, (err, res) => {
                    if (err) {
                        err.message = `[${key}] ${err.message}`;
                        return reject(err);
                    }
                    files[key].contents = Buffer.from(res);
                    resolve(true);
                });
            });
        });
        Promise.all(promises).then(() => done(null, files, metalsmith));
    };
}
function loadRepository(tpl) {
    return new Promise((resolve, reject) => {
        const dest = tmp_1.default.dirSync().name;
        if ((0, utils_2.isGitUrl)(tpl)) {
            const spinner = (0, nanospinner_1.createSpinner)("downloading template").start();
            (0, utils_1.downloadRepo)(tpl, dest);
            spinner.success();
            resolve(dest);
        }
        else if (fs_1.default.existsSync(tpl)) {
            copyTo(tpl, dest, () => {
                resolve(dest);
            });
        }
        else {
            reject(new Error("unknown template path"));
        }
    });
}
async function resolveOption(opts) {
    const questions = [
        {
            type: "input",
            message: "folder name",
            name: "projectName",
            validate: function (val) {
                const reg = /[a-zA-Z0-9\-_]+/;
                if (!val) {
                    return "please input your project name";
                }
                else if (reg.test(val)) {
                    return true;
                }
                else {
                    return `input should be ${reg.toString}`;
                }
            },
        },
    ];
    const tplsMap = (0, utils_1.getTemplateRecords)();
    if (!opts.templatePath) {
        const tplNames = Object.keys(tplsMap);
        if (!tplNames.length) {
            logger_1.default.warn(`
       can not found any template
       try to add template by run
       $ inix add`);
            return;
        }
        questions.unshift({
            type: "list",
            name: "template",
            message: "select template which you want",
            default: tplNames[0],
            choices: tplNames,
        });
    }
    const answers = await inquirer_1.default.prompt(questions);
    opts.data = answers;
    let tplInfo = tplsMap[answers.template];
    let templatePath;
    if (opts.templatePath) {
        templatePath = opts.templatePath;
    }
    else if (tplInfo) {
        templatePath = tplInfo.templatePath;
        if ((0, utils_2.isGitUrl)(templatePath) && tplInfo.branch) {
            templatePath += `#${tplInfo.branch}`;
        }
    }
    opts.templatePath = templatePath;
    return opts;
}
async function default_1(opts) {
    opts = await resolveOption(opts);
    if (!opts)
        return;
    opts.shouldAskQuestions = true;
    await createApp(opts);
}
exports.default = default_1;
async function createApp(opts) {
    const projectTemplatePath = await loadRepository(opts.templatePath);
    Object.assign(opts, { projectTemplatePath });
    await initProject(opts);
}
exports.createApp = createApp;
